import unittest
from pathlib import Path

from verminator.releasemeta import ProductReleaseMeta
from verminator.utils import *
from verminator.verminator import VersionedInstance


class VersionedInstanceCase(unittest.TestCase):

    def setUp(self):
        this_file = Path(__file__)
        self.tdc3ex_yml = this_file.parent.joinpath('releasesmeta/tdc3ex.yml')
        self.versioned_instance_yml = this_file.parent.joinpath('releasesmeta/versioned_instance.yml')

    def test_official(self):
        meta = ProductReleaseMeta(self.tdc3ex_yml)
        versioned_instance = VersionedInstance(**yaml.load(open(self.versioned_instance_yml)))
        self.assertTrue(str(versioned_instance.min_tdc_version) == 'tdc-2.0.0-rc0')
        self.assertTrue(str(versioned_instance.max_tdc_version) == 'tdc-2.0.0-rc9')
        versioned_instance._validate_releases(meta)
        self.assertTrue(versioned_instance is not None)
        release = versioned_instance.get_release('5.2.2')
        for dep in release.dependencies:
            self.assertTrue(str(release.dependencies[dep][0]) == 'transwarp-5.2.2-final')
            self.assertTrue(str(release.dependencies[dep][1]) == 'transwarp-5.2.2-final')



import unittest
from pathlib import Path

from verminator.releasemeta import ProductReleaseMeta
from verminator.utils import *


class ProductReleaseMetaCase(unittest.TestCase):

    def setUp(self):
        this_file = Path(__file__)
        self.oem_yml = this_file.parent.joinpath('releasesmeta/oem.yml')
        self.tdc_yml = this_file.parent.joinpath('releasesmeta/tdc.yml')
        self.tdc2ex_yml = this_file.parent.joinpath('releasesmeta/tdc2ex.yml')
        self.tdc3ex_yml = this_file.parent.joinpath('releasesmeta/tdc3ex.yml')

    def test_official(self):
        meta = ProductReleaseMeta(self.tdc_yml)
        releases = [str(i) for i in meta.get_major_versioned_releases().keys()]
        self.assertTrue('tdc-1.0' in releases)
        self.assertTrue('tdc-1.1' in releases)
        self.assertTrue('tdc-1.2' in releases)
        self.assertFalse('tdc-1.1.0-final' in releases)

    def test_oem(self):
        meta = ProductReleaseMeta(self.oem_yml)
        releases = [str(i) for i in meta.get_major_versioned_releases().keys()]
        self.assertTrue('gzes-1.0' in releases)
        self.assertTrue('gzes-1.1' in releases)
        self.assertTrue('gzes-1.2' in releases)
        self.assertFalse('gzes-1.1.0-final' in releases)

    def assert_vrange_equal(self, vr, other):
        vr = (parse_version(vr[0]), parse_version(vr[1]))
        other = (parse_version(other[0]), parse_version(other[1]))
        self.assertTrue(vr[0] == other[0] and vr[1] == other[1])

    def test_get_tdc_version_range_legacy(self):
        meta = ProductReleaseMeta(self.tdc_yml)

        pv = meta.get_compatible_versions('sophonweb-1.3.0-final')
        self.assert_vrange_equal(pv.get('tdc')[0], ('tdc-1.1.0-rc2', 'tdc-1.2.1-rc1'))
        self.assert_vrange_equal(pv.get('transwarp')[0], ('transwarp-5.2.1-final', 'transwarp-5.2.3-final'))
        self.assert_vrange_equal(pv.get('sophonweb')[0], ('sophonweb-1.3.0-final', 'sophonweb-1.3.0-final'))

        pv = meta.get_tdc_version_range('sophonweb-1.3.0-final')
        self.assert_vrange_equal(pv, ('tdc-1.1.0-rc2', 'tdc-1.2.1-rc1'))

        pv = meta.get_tdc_version_range('sophonweb-2.1.0-final')
        self.assert_vrange_equal(pv, ('tdc-1.2.0-rc3', 'tdc-1.2.1-rc1'))

        pv = meta.get_tdc_version_range()
        self.assert_vrange_equal(pv, ('tdc-1.0.0-rc1', 'tdc-1.2.1-rc1'))

    def test_get_compatible_versions(self):
        meta = ProductReleaseMeta(self.tdc2ex_yml)
        pv = meta.get_compatible_versions('tdc-2.0.0-rc1')
        self.assert_vrange_equal(pv.get('transwarp')[0], ('transwarp-5.2.1-final', 'transwarp-6.0.1-final'))
        self.assert_vrange_equal(pv.get('sophonweb')[0], ('sophonweb-1.3.0-final', 'sophonweb-2.2.0-final'))
        self.assert_vrange_equal(pv.get('tos')[0], ('tos-1.9.0-final', 'tos-1.9.1-final'))
        self.assert_vrange_equal(pv.get('tdc')[0], ('tdc-2.0.0-rc1', 'tdc-2.0.0-rc1'))

        pv = meta.get_compatible_versions('tdc-2.0.0-rc3')
        self.assert_vrange_equal(pv.get('transwarp')[0], ('transwarp-5.2.1-final', 'transwarp-6.0.2-final'))
        self.assert_vrange_equal(pv.get('sophonweb')[0], ('sophonweb-2.2.0-final', 'sophonweb-2.2.1-final'))
        self.assert_vrange_equal(pv.get('tos')[0], ('tos-1.9.2-final', 'tos-1.9.2-final'))

        pv = meta.get_compatible_versions('sophonweb-2.2.0-final')
        self.assert_vrange_equal(pv.get('transwarp')[0], ('transwarp-5.2.1-final', 'transwarp-5.2.3-final'))
        self.assert_vrange_equal(pv.get('tdc')[0], ('tdc-2.0.0-rc0', 'tdc-2.0.0-rc3'))
        self.assert_vrange_equal(pv.get('sophonweb')[0], ('sophonweb-2.2.0-final', 'sophonweb-2.2.0-final'))

        pv = meta.get_compatible_versions('transwarp-5.2.1-final')
        self.assert_vrange_equal(pv.get('transwarp')[0], ('transwarp-5.2.1-final', 'transwarp-5.2.1-final'))
        self.assert_vrange_equal(pv.get('tdc')[0], ('tdc-2.0.0-rc0', 'tdc-2.0.0-rc3'))
        self.assert_vrange_equal(pv.get('sophonweb')[0], ('sophonweb-1.3.0-final', 'sophonweb-2.2.1-final'))

        pv = meta.get_compatible_versions('transwarp-6.0.1-final')
        self.assert_vrange_equal(pv.get('tdc')[0], ('tdc-2.0.0-rc0', 'tdc-2.0.0-rc3'))
        self.assert_vrange_equal(pv.get('sophonweb')[0], ('sophonweb-1.3.0-final', 'sophonweb-2.2.1-final'))

        pv = meta.get_compatible_versions('transwarp-6.0.2-final')
        self.assert_vrange_equal(pv.get('tdc')[0], ('tdc-2.0.0-rc3', 'tdc-2.0.0-rc3'))
        self.assert_vrange_equal(pv.get('sophonweb')[0], ('sophonweb-2.2.0-final', 'sophonweb-2.2.1-final'))

    def test_get_tdc_version_range(self):
        meta = ProductReleaseMeta(self.tdc2ex_yml)
        pv = meta.get_tdc_version_range('sophonweb-1.3.0-final')
        self.assert_vrange_equal(pv, ('tdc-2.0.0-rc0', 'tdc-2.0.0-rc1'))

        pv = meta.get_tdc_version_range('sophonweb-1.4.0-final')
        self.assert_vrange_equal(pv, ('tdc-2.0.0-rc0', 'tdc-2.0.0-rc1'))

        pv = meta.get_tdc_version_range('sophonweb-2.2.0-final')
        self.assert_vrange_equal(pv, ('tdc-2.0.0-rc0', 'tdc-2.0.0-rc3'))

        pv = meta.get_tdc_version_range('sophonweb-2.2.1-final')
        self.assert_vrange_equal(pv, ('tdc-2.0.0-rc3', 'tdc-2.0.0-rc3'))

        pv = meta.get_tdc_version_range()
        self.assert_vrange_equal(pv, ('tdc-2.0.0-rc0', 'tdc-2.0.0-rc3'))

        pv = meta.get_tdc_version_range('sophonweb-2.1')
        self.assert_vrange_equal(pv, ('tdc-2.0.0-rc0', 'tdc-2.0.0-rc1'))

        pv = meta.get_tdc_version_range('sophonweb-2.2')
        self.assert_vrange_equal(pv, ('tdc-2.0.0-rc0', 'tdc-2.0.0-rc3'))

    def test_get_compatible_versions_for_instance(self):
        meta = ProductReleaseMeta(self.tdc3ex_yml)
        pv = meta.get_compatible_versions('5.2.2', instance_name='tdh-metrics-exporter')
        self.assert_vrange_equal(pv.get('transwarp')[0], ('transwarp-5.2.2-final', 'transwarp-5.2.2-final'))
        self.assertTrue(pv.get('sophonweb') is None)
        self.assertTrue(pv.get('tdc') is None)
        self.assertTrue(pv.get('tos') is None)

        pv = meta.get_compatible_versions('6.0.0', instance_name='tdh-metrics-exporter')
        self.assert_vrange_equal(pv.get('transwarp')[0], ('transwarp-6.0.2-final', 'transwarp-6.0.2-final'))
        self.assertTrue(pv.get('sophonweb') is None)
        self.assertTrue(pv.get('tdc') is None)
        self.assertTrue(pv.get('tos') is None)

        pv = meta.get_compatible_versions('sophonweb-2.2.1-final', instance_name='workflow')
        self.assert_vrange_equal(pv.get('transwarp')[0], ('transwarp-5.2.4-final', 'transwarp-5.2.4-final'))
        self.assert_vrange_equal(pv.get('sophonweb')[0], ('sophonweb-2.2.1-final', 'sophonweb-2.2.1-final'))
        self.assert_vrange_equal(pv.get('tos')[0], ('tos-1.9.2-final', 'tos-1.9.2-final'))
        self.assert_vrange_equal(pv.get('tdc')[0], ('tdc-2.0.0-rc3', 'tdc-2.0.0-rc3'))



import unittest

from verminator.utils import *


class UtilsCase(unittest.TestCase):

    def setUp(self):
        pass

    def assert_vrange_equal(self, vr, other):
        vr = (parse_version(vr[0]), parse_version(vr[1]))
        other = (parse_version(other[0]), parse_version(other[1]))
        self.assertTrue(vr[0] == other[0] and vr[1] == other[1])

    def test_filter_vrange(self):
        vr = (parse_version('sophonweb-1.3.0-final'), parse_version('sophonweb-2.2.1-final'))
        other = (parse_version('sophonweb-2.2.0-final'), parse_version('sophonweb-2.2.0-final'))
        filtered = filter_vrange(vr, other)
        self.assert_vrange_equal(filtered, other)

        vr = (parse_version('sophonweb-1.3.0-final'), parse_version('sophonweb-1.3.0-final'))
        other = (parse_version('sophonweb-1.3.0-final'), parse_version('sophonweb-1.3.0-final'))
        self.assert_vrange_equal(filter_vrange(vr, other), other)

        vr = (parse_version('sophonweb-2.2.0-final'), parse_version('sophonweb-2.2.0-final'))
        other = (parse_version('sophonweb-1.3.0-final'), parse_version('sophonweb-1.3.0-final'))
        self.assertTrue(filter_vrange(vr, other) is None)

        vr = (parse_version('sophonweb-2.0.0-final'), parse_version('sophonweb-2.0.0-final'))
        other = (parse_version('sophonweb-1.3.0-final'), parse_version('sophonweb-2.2.0-final'))
        self.assert_vrange_equal(filter_vrange(vr, other), vr)

        vr = (parse_version('transwarp-5.2.1-final'), parse_version('transwarp-6.0.2-final'))
        other = (parse_version('transwarp-5.2.1-final'), parse_version('transwarp-5.2.1-final'))
        self.assert_vrange_equal(filter_vrange(vr, other), other)

    def test_parse_version(self):
        v = parse_version('tdc-1.2')
        self.assertTrue(v.maintenance is None and v.build is None and v.suffix_version is None)
        v = parse_version('tdc-1.2.1-final')
        self.assertTrue(v.suffix == 'final' and v.minor == 2 and v.maintenance == 1)

    def test_is_major_version(self):
        self.assertTrue(is_major_version('tdc-1.2'))
        self.assertTrue(not is_major_version('tdc-1.2.1'))
        self.assertTrue(not is_major_version('tdc-1.2.1-final'))

    def test_replace_product_name(self):
        self.assertTrue(str(replace_product_name('tdc-1.2', 'gzes')) == 'gzes-1.2')
        self.assertTrue(str(replace_product_name('transwarp-1.2', 'gzes', by='tdc')) == 'transwarp-1.2')
        self.assertTrue(str(replace_product_name('tdc-1.2', 'gzes', by='tdc')) == 'gzes-1.2')

    def test_concatenate_vranges(self):
        cvr = concatenate_vranges([
            (parse_version('sophonweb-1.2.0-final'), parse_version('sophonweb-2.2.0-final')),
            (parse_version('sophonweb-1.3.0-rc0'), parse_version('sophonweb-1.3.0-rc3'))
        ])
        self.assertTrue(len(cvr) == 1)
        self.assert_vrange_equal(cvr[0], (parse_version('sophonweb-1.2.0-final'), parse_version('sophonweb-2.2.0-final')))

        cvr = concatenate_vranges([
            (parse_version('tdc-1.0.0-rc0'), parse_version('tdc-1.0.0-final')),
            (parse_version('tdc-1.1.0-rc0'), parse_version('tdc-1.1.0-final')),
            (parse_version('tdc-1.2.0-rc0'), parse_version('tdc-1.2.0-final')),
        ])
        self.assertTrue(len(cvr) == 1)
        self.assert_vrange_equal(cvr[0], (parse_version('tdc-1.0.0-rc0'), parse_version('tdc-1.2.0-final')))

        cvr = concatenate_vranges([
            (parse_version('tdc-1.0.0-rc0'), parse_version('tdc-1.0.0-final')),
            (parse_version('transwarp-1.1.0-rc0'), parse_version('transwarp-1.1.0-final')),
            (parse_version('tdc-1.2.0-rc0'), parse_version('tdc-1.2.0-final')),
        ])
        self.assertTrue(len(cvr) == 2)
        self.assert_vrange_equal(cvr[0], (parse_version('tdc-1.0.0-rc0'), parse_version('tdc-1.2.0-final')))
        self.assert_vrange_equal(cvr[1], (parse_version('transwarp-1.1.0-rc0'), parse_version('transwarp-1.1.0-final')))


# Meta data for release info in yaml, e.g.,
# **********************
# Releases:
# - products:
#   - {max: tos-1.8.0.1, min: tos-1.8.0.1}
#   - {max: transwarp-5.1.0-final, min: transwarp-5.1.0-final}
#   release_name: tdc-1.0.0-rc1
# - products:
#   - {max: tos-1.8.0-rc2, min: tos-1.8.0-rc2}
#   - {max: transwarp-5.1.0-final, min: transwarp-5.1.0-final}
#   release_name: tdc-1.0.0-rc2
# ************************
from .config import verminator_config as VC
from .utils import *

__all__ = ['ProductReleaseMeta']


class ProductReleaseMeta(object):
    """ Processing `releases_meta.yaml`.
    """

    DEFAULT_INSTANCE_NAME = None

    def __init__(self, yaml_file):
        with open(yaml_file) as ifile:
            self._raw_data = yaml.load(ifile, Loader=yaml.FullLoader)
        # -----------------------------------------------------------
        # Hierarchical version constraints, including:
        # * TDC release constraints on other product versions;
        # * Product release constraints on others else;
        # * Instance release constraints on other product versions;
        #
        # Inner data structure:
        # * {instance_name: release_ver: product_line: (minv, maxv)}
        # -----------------------------------------------------------
        self._releases = self._load_releases()
        self._major_versioned_releases = self._load_releases(True)

    def _load_releases(self, major_versioned=False):
        """ Read releases meta info of product lines
        """
        all_releases = dict()

        for r in self._raw_data.get('Releases', list()):

            # Extract release meta info.
            instance_name = r.get('instance', self.DEFAULT_INSTANCE_NAME)
            release_ver = parse_version(r.get('release_name'), major_versioned)
            if instance_name not in all_releases:
                all_releases[instance_name] = dict()
            if release_ver not in all_releases[instance_name]:
                all_releases[instance_name][release_ver] = dict()

            versioned_releases = all_releases[instance_name][release_ver]

            # Extract product constraints for each release version
            products = r.get('products', list())
            for p in products:
                minv = parse_version(p.get('min'), major_versioned)
                maxv = parse_version(p.get('max'), major_versioned)
                minv_name = product_name(minv)
                maxv_name = product_name(maxv)
                assert minv_name == maxv_name, \
                    'Product version should have the same prefix name: %s vs. %s' \
                    % (minv_name, maxv_name)

                pname = product_name(minv)
                if pname not in versioned_releases:
                    versioned_releases[pname] = (minv, maxv)
                else:
                    vrange = versioned_releases[pname]
                    versioned_releases[pname] = concatenate_vranges(
                        [vrange, (minv, maxv)],
                        hard_merging=major_versioned
                    )[0]

        return all_releases

    def get_releases(self, instance_name=None):
        """
        Get all versioned releases for specific instance_name.

        :param instance_name: the specific instance name or None if not present.
        :return: {release_ver: product_line: (minv, maxv)} or empty dict.
        """
        return self._releases.get(instance_name, dict())

    def get_major_versioned_releases(self, instance_name=None):
        """
        Get all major versioned releases for specific instance_name.

        :param instance_name: the specific instance name or None if not present.
        :return: {release_ver: product_line: (minv, maxv)} or empty dict.
        """
        return self._major_versioned_releases.get(instance_name, dict())

    def get_tdc_version_range(self, version=None, instance_name=None):
        """Given a specific product version,
        :return: the compatible tdc (complete) version range, (minv, maxv)
        """
        tdc_versions = [i for i in self.get_releases(instance_name).keys()
                        if product_name(i) == VC.OEM_NAME]
        sorted_tdc_version = sorted(tdc_versions, key=cmp_to_key(
            lambda x, y: x.compares(y)
        ))

        rv1 = rv2 = None
        if version is None:
            rv1, rv2 = sorted_tdc_version[0], sorted_tdc_version[-1]
        else:
            # Get compatible tdc versions in a normalized way
            version = parse_version(version)

            # Get product-versions mapping
            pvmap = self.get_compatible_versions(version, True, instance_name)

            versions = list()  # [(minv, maxv)]
            for v1, v2 in pvmap.get(VC.OEM_NAME, list()):
                if is_major_version(v1):
                    minv, maxv = None, None
                    for v in sorted_tdc_version:
                        if v1 == to_major_version(v):
                            minv = v
                            break
                    for v in sorted_tdc_version[::-1]:
                        if v2 == to_major_version(v):
                            maxv = v
                            break
                    if None in (minv, maxv):
                        raise ValueError('Can not get valid tdc version range for {}'.format(version))
                    versions.append((minv, maxv))
                else:
                    versions.append((v1, v2))

            if len(versions) > 0:
                rv1, rv2 = versions[0][0], versions[-1][1]

        return None if None in (rv1, rv2) else (rv1, rv2)

    def get_compatible_versions(self, version, minor_versioned=False, instance_name=None):
        """
        Given a specific product version. If the instance_name is present,
        more constraints on the instance would be considered.
        Otherwise, the instance-specific constraints would be ignored.

        :param version: the product version.
        :param minor_versioned: if checking minor versions only.
        :param instance_name: the specified instance name.
        :return: the compatible product version ranges, {product: [(minv, maxv)]}
        """
        version = parse_version(version)
        product = product_name(version)

        # Check that the version is complete or in major form
        _is_major_version = is_major_version(version)
        instance_releases = dict()
        default_releases = dict()

        # Prepare instance-specific and default release meta info.
        if not _is_major_version or minor_versioned:
            default_releases = self.get_releases()
        else:
            default_releases = self.get_major_versioned_releases()

        if instance_name is not None:
            if not _is_major_version or minor_versioned:
                instance_releases = self.get_releases(instance_name=instance_name)
            else:
                instance_releases = self.get_major_versioned_releases(instance_name=instance_name)

        # Unify instance-specific and default release meta info, if present
        releases = default_releases
        for r, product_versions in instance_releases.items():
            if r not in releases:
                releases[r] = product_versions
            else:
                default_product_versions = releases[r]
                for p, vrange in product_versions.items():
                    # Filter default product version range with declared ones, forcedly
                    filtered = filter_vrange(default_product_versions[p], vrange)

                    assert filtered is not None, \
                        'Warning: version declaration conflicts for {}: {} and {}, please fix releases_meta.yml' \
                            .format(r, default_product_versions[p], vrange)

                    default_product_versions[p] = filtered

        # Differentiate derived and declared constraints
        derived_constraints = {}
        declared_constraints = {
            product: [(version, version)]  # We treat the input as declared constraint
        }

        def add_constraint(store, product, vrange):
            if product not in store:
                store[product] = list()
            store[product].append(vrange)

        # Construct constraints
        for r, product_versions in releases.items():
            pname = product_name(r)

            # Classify candidates as DECLARED if the target product is in the releases,
            # otherwise as DERIVED.
            if product == pname:  # DECLARED
                if r == version:  # Matched declared version
                    # Update constraints
                    add_constraint(declared_constraints, pname, (r, r))
                    for p, vrange in product_versions.items():
                        add_constraint(declared_constraints, p, vrange)

            else:  # DERIVED
                # Omit mismatched product-line
                # We assume the derived are concluded from EXPLICIT dependencies.
                # Then we always have no derived constraints for third-party instances,
                # because instance-specific constraints are not supported yet.
                if product not in product_versions:
                    continue

                if not _is_major_version:
                    if not version.in_range(product_versions[product][0], product_versions[product][1]):
                        continue  # Omit version range not containing target minor version
                else:
                    vmin, vmax = product_versions[product]
                    if not version.in_range(to_major_version(vmin), to_major_version(vmax)):
                        continue

                # Update constraints
                add_constraint(derived_constraints, pname, (r, r))
                for p, vrange in product_versions.items():
                    add_constraint(derived_constraints, p, vrange)

        # Merging constraints
        merged = self._merge_dd(declared_constraints, derived_constraints, _is_major_version)

        return merged

    def _merge_dd(self, declared_constraints, derived_constraints, hard_merging=False):
        """
        Merge both DECLARED and DERIVED constraints.

        In the merging algorithm, the DECLARED releasesz represent strong
        dependencies between product lines, which takes more priority
        than the DERIVED ones.

        :param declared_constraints: {product: [(minv, maxv)]}
        :param derived_constraints: {product: [(minv, maxv)]}
        :return: the merged product version ranges, {product: [(minv, maxv)]}
        """
        merged = dict()
        keys = set(list(declared_constraints.keys()) + list(derived_constraints.keys()))
        for k in keys:
            declared = derived = None
            if k in declared_constraints:
                declared = concatenate_vranges(declared_constraints[k], hard_merging)
            if k in derived_constraints:
                derived = concatenate_vranges(derived_constraints[k], hard_merging)

            if derived is None and declared is None:
                continue
            elif derived is not None and declared is None:
                merged[k] = derived
            elif derived is None and declared is not None:
                merged[k] = declared
            else:
                # Filter derived by declared
                merged[k] = list()
                for v in derived:
                    for w in declared:
                        f = filter_vrange(v, w)
                        if f is not None:
                            v = f
                    if v is not None:
                        merged[k].append(v)
        return merged



#!/usr/bin/env python3
# Module stolen from product-meta:
# http://172.16.1.41:10080/TDC/product-meta/blob/tdc-1.2/tests/validate_instance_images.py
from pathlib import Path

import yaml
from flex_version import FlexVersion


class ReleaseDep(object):
    def __init__(self, dep_desc):
        self.type = dep_desc['type']
        self.max_version = dep_desc['max-version']
        self.min_version = dep_desc['min-version']


class ApplicationDep(object):
    def __init__(self, name, module_name, ori_version):
        self.type = module_name
        self.name = name
        self.version = ori_version


class Application(object):
    def __init__(self, instance, name):
        self.dependencies = list()
        self.name = name
        self.type = instance.get('moduleName')
        self.version = instance.get('version')
        if instance.get('dependencies') is not None:
            for dependency in instance.get('dependencies'):
                self.dependencies.append(ApplicationDep(name, dependency.get('moduleName'), self.version))


class Product(object):
    def __init__(self, product, json_path, version):
        self.component = list()
        name = str(json_path).split('products/')[1]
        name = name.split('/')[0]
        self.name = name
        self.edition = version
        components = product.get('components')
        for component in components:
            name = component.split('-')[0].upper()
            version = component.split('-')[1].lower()
            self.component.append(Component(name, version))


class Component(object):
    def __init__(self, name, version):
        self.name = name
        self.version = version


class ReleaseInfo(object):
    # All metainfo of versioned instances:
    # {instance: {version: ReleaseInfo}}
    __instance_releases = dict()

    def __init__(self, instance_name, release_version, is_final, instance_version, dependencies=None):
        self.instance_name = instance_name
        self.release_version = release_version
        self.is_final = is_final
        self.instance_version = instance_version
        if dependencies:
            self.dependencies = [ReleaseDep(i) for i in dependencies]
        else:
            self.dependencies = list()

    @classmethod
    def add_release(cls, instance_name, release_desc, instance_version):
        release_version = release_desc['release-version']
        is_final = release_desc['final']
        image_version = release_desc['image-version']
        dep_desc = release_desc['dependencies']

        release = ReleaseInfo(instance_name, release_version, is_final, instance_version, dep_desc)

        if instance_name not in cls.__instance_releases:
            cls.__instance_releases[instance_name] = dict()
        if release.release_version not in cls.__instance_releases[instance_name]:
            cls.__instance_releases[instance_name][release.release_version] = dict()
        else:
            raise ValueError('Duplicated release version {} for {}'.format(
                release.release_version, instance_name)
            )
        cls.__instance_releases[instance_name][release.release_version] = release

        return release

    @classmethod
    def all_instance_releases(cls):
        return cls.__instance_releases


def scan_instances(root_dir, omitsample=False):
    """
    Scan all instances directories
    """
    rp = Path(root_dir)
    for instance in rp.iterdir():
        if not instance.is_dir() or \
            (omitsample and instance.name.startswith('_')):
            continue
        instance = instance.name
        inspath = Path(rp.joinpath(instance))
        versions = [x for x in inspath.iterdir() if x.is_dir()]
        for version in versions:
            version = version.name
            vpath = inspath.joinpath(version)
            imgpath = vpath.joinpath('images.yaml')
            if not imgpath.exists():
                # Omit subfolder without valid images yaml
                continue
            images = yaml.load(open(imgpath), Loader=yaml.FullLoader)

            # Validate images meta info
            validate_versioned_image(images, instance, version)


def validate_versioned_image(images, instance_name, instance_version):
    """
    Validate the meta info of images defined for each instance
    """
    print('Validating versioned {}, {}'.format(instance_name, instance_version))
    assert images['instance-type'].lower() == instance_name.lower()
    assert images['major-version'] == instance_version
    assert 'images' in images
    assert 'releases' in images
    assert 'min-tdc-version' in images
    assert 'hot-fix-ranges' in images

    # Collect all releases of instances
    releases = dict()
    for r in images.get('releases', list()):
        release_info = ReleaseInfo.add_release(instance_name, r, instance_version)
        releases[release_info.release_version] = release_info

    # Validate hot-fix range: each defined release should be in a hot-fix range
    hot_fixes = images.get('hot-fix-ranges', list())
    for rv in releases:
        found = False
        for fix_range in hot_fixes:
            if FlexVersion.in_range(rv, minv=fix_range['min'], maxv=fix_range['max']):
                found = True
        assert found, 'Release version {} of {} not in a valid hot-fix range' \
            .format(rv, instance_name)

    # Validate dependence min-max range: min <= max
    for release_info in releases.values():
        for dep in release_info.dependencies:
            res = FlexVersion.compares(dep.min_version, dep.max_version)
            assert res <= 0, 'Invalid min-max range [min: {}, max: {}] for version {} of {}' \
                .format(dep.min_version, dep.max_version, release_info.release_version, instance_name)


def _find_a_ranged_version(instance_name, minv, maxv):
    """
    Given a version range, check if a valid version is defined in the range.
    """
    all_instance_releases = ReleaseInfo.all_instance_releases()
    versions = all_instance_releases.get(instance_name, dict())
    found = False
    for v, release_info in versions.items():
        found = FlexVersion.in_range(v, minv, maxv)
        if found:
            break
    return found


def validate_dependence_versions():
    """
    Each dependence should have at least a version defined.
    """
    print('Total {} instances defined'.format(len(ReleaseInfo.all_instance_releases())))
    all_instance_releases = ReleaseInfo.all_instance_releases()

    # Validate dependencies: each dependence should have a valid
    # version as in defined range
    for instance_name, versions in all_instance_releases.items():
        for instance_version, release_info in versions.items():
            print('Validating dependence of {} version {}'.format(
                release_info.instance_name, instance_version))
            dependencies = release_info.dependencies
            for dep in dependencies:
                dep_type = dep.type
                minv = dep.min_version
                maxv = dep.max_version

                assert dep_type in all_instance_releases, \
                    "No dependence found {} for version {} of {}" \
                        .format(dep_type, instance_version, instance_name)

                found = _find_a_ranged_version(dep_type, minv=minv, maxv=maxv)
                assert found, "No valid version defined for {} with max: {}, min: {}" \
                    .format(dep_type, maxv, minv)



__all__ = ['verminator_config']
import os


class SingletonMetaClass(type):
    """ A metaclass that creates a Singleton base class when called. """
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(SingletonMetaClass, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class Singleton(SingletonMetaClass('SingletonMeta', (object,), {})):
    pass


class VerminatorConfig(Singleton):
    _OEM_ORIGIN = 'tdc'
    OEM_NAME = 'tdc'

    def set_oem(self, oemname):
        if oemname is not None:
            self.OEM_NAME = oemname
        elif os.getenv('OEM_NAME', ''):
            self.OEM_NAME = os.getenv('OEM_NAME')


verminator_config = VerminatorConfig()



#!/usr/bin/env python3
# This library deals with the instances images of product-meta,
# which has following folder structure as mapped by the data objects:
# product-meta:
# |__ ...
# |__instances [class Instance]
#    |__version1 [class VersionedInstance]
#       |__images.yml [class Release]
from pathlib import Path

from .config import verminator_config as VC
from .utils import *

__all__ = ['Instance', 'VersionedInstance', 'Release']


class Instance(object):
    def __init__(self, instance_type, instance_folder, omitsample=False):
        self.instance_type = instance_type
        self.instance_folder = Path(instance_folder)
        self.versioned_instances = dict()  # {major_version_num: VersionedInstance}

        if omitsample and instance_type.startswith('_'):
            # Omit instance with private symbol '_'
            return

        for ver in self.instance_folder.iterdir():
            if omitsample and ver.name.startswith('_'):
                # Omit instance version with private symbol '_'
                continue
            image_file = ver.joinpath('images.yaml')
            if not image_file.exists():
                # Omit subfolder without valid images yaml
                continue
            dat = yaml.load(open(image_file), Loader=yaml.FullLoader)
            ins = VersionedInstance(**dat)
            self.add_versioned_instance(ver.name, ins)

    def add_versioned_instance(self, major_version_num, instance):
        assert major_version_num not in self.versioned_instances, \
            'Duplicated version %s for instance %s' % (major_version_num, self.instance_type)
        self.versioned_instances[major_version_num] = instance

    def get_versioned_instance(self, major_version_num):
        return self.versioned_instances.get(major_version_num, None)

    def create_release(self, version):
        """Create a new release and add to appropriate VersionedInstance.

        The version is preferentially added to VersionedInstance containing its
        corresponding major version. Otherwise a new VersionedInstance created.
        """
        version = parse_version(version)
        major_version_num = '{}.{}'.format(version.major, version.minor)

        # Find VersionedInstance which contains the major version
        ref_instance = None
        if major_version_num in self.versioned_instances:
            ref_instance = self.versioned_instances[major_version_num]
        else:
            major_version = to_major_version(version)  # with prefix say tdc-2.0
            for versioned_ins in self.versioned_instances.values():
                if versioned_ins.has_release(major_version):
                    ref_instance = versioned_ins
                    break

        if ref_instance is not None:
            ref_instance.create_release(version, None, True)
        else:
            # Find the latest release with the same product as reference
            ref_release = None
            for ver_num in sorted(self.versioned_instances.keys(), reverse=True):
                ins = self.versioned_instances[ver_num]
                latest_r = ins.find_latest_release(product_name(version))
                if latest_r is not None:
                    ref_release = latest_r
                    ref_instance = ins
                if ref_instance is not None:
                    break

            if ref_instance is None:
                raise ValueError('No valid final reference release found for creating {} of {}'.format(
                    version, ref_instance.instance_type
                ))

            new_instance = copy.deepcopy(ref_instance)
            new_instance.major_version = major_version_num
            new_instance._hot_fix_ranges = list()
            new_instance._releases = dict()
            new_instance.create_release(version, ref_release, with_major=True)
            self.versioned_instances[major_version_num] = new_instance

    def has_release(self, release_version):
        for ver, versioned_ins in self.versioned_instances.items():
            if versioned_ins.has_release(release_version):
                return True
        return False

    def validate(self, release_meta):
        self._validate_declared_tdc_releases(release_meta)
        for ver, versioned_ins in self.versioned_instances.items():
            print(self.instance_folder.joinpath(ver))
            # Validate specific versioned instance
            versioned_ins.validate(release_meta)

    def _validate_declared_tdc_releases(self, release_meta):
        """The declared TDC releases in release_meta.yaml should be also declared in images.yaml
        """
        tdc_releases = [i for i in release_meta.get_releases().keys() if i.prefix == VC.OEM_NAME]

        # Check if the instance contains TDC versioned releases
        found = False
        for ver, versioned_ins in self.versioned_instances.items():
            for release in versioned_ins.releases:
                if release.release_version.prefix == VC.OEM_NAME:
                    found = True
                    break

        if found:
            for tdc_version in tdc_releases:
                if not self.has_release(tdc_version):
                    raise ValueError("Absent TDC release '{}' defined in meta for instance {}".format(
                        tdc_version, self.instance_type
                    ))

    def dump(self):
        for ver, ins in self.versioned_instances.items():
            version_folder = self.instance_folder.joinpath(ver)
            if not version_folder.exists():
                version_folder.mkdir(parents=True)
            image_file = version_folder.joinpath('images.yaml')
            yaml_str = ins.to_yaml()
            if yaml_str:
                with open(image_file, 'w') as of:
                    of.write(yaml_str)


class VersionedInstance(object):
    """A versioned instance
    """

    def __init__(self, **kwargs):
        self.instance_type = kwargs.get('instance-type')

        # The major version corresponds to chart version in helm.
        self.major_version = parse_version(kwargs.get('major-version'))

        self._min_tdc_version = parse_version(
            kwargs.get('min-tdc-version', None)
        )

        self._max_tdc_version = parse_version(
            kwargs.get('max-tdc-version', None))

        if self._min_tdc_version.compares(self._max_tdc_version) > 0:
            raise ValueError('Invalid min-max tdc version range for %s %s' %
                             (self.instance_type, self.major_version))

        self._hot_fix_ranges = list()  # [(minv, maxv)]
        for item in kwargs.get('hot-fix-ranges', list):
            self.add_hot_fix_range(item.get('min'), item.get('max'))

        self._images = dict()  # {var: { name:name, role: role, roles: []}}
        for item in kwargs.get('images', dict()):
            self.add_image(item)

        self._releases = dict()  # {release_ver: Release}
        for release in kwargs.get('releases', dict()):
            self.add_release(release)

    @property
    def min_tdc_version(self):
        """Get the """
        assert not is_major_version(self._min_tdc_version), \
            'The min tdc version should be a valid final version'
        return self._min_tdc_version

    @property
    def max_tdc_version(self):
        assert not is_major_version(self._max_tdc_version), \
            'The max tdc version should be a valid final version'
        return self._max_tdc_version

    @property
    def hot_fix_ranges(self):
        """Get a list of hot-fix ranges in form [(minv, maxv)]
        """
        return self._hot_fix_ranges

    @property
    def images(self):
        """Get a dict of image variables and names
        """
        return self._images.items()

    @property
    def releases(self):
        """Get a list of releases
        """
        return self._releases.values()

    @property
    def ordered_releases(self):
        """ Get a list of ordered releases by versions.
        """
        return sorted(self._releases.values(), key=cmp_to_key(
            lambda x, y: x.release_version.compares(y.release_version)
        ))

    def add_hot_fix_range(self, _min, _max):
        """Add a new hot-fix range to VersionedInstance from raw data.
        """
        if isinstance(_min, str):
            _min = parse_version(_min)
        if isinstance(_max, str):
            _max = parse_version(_max)

        if _min.compares(_max) > 0:
            raise ValueError(
                'Invalid hot fix range for %s %s' %
                (self.instance_type, self.major_version)
            )

        self._hot_fix_ranges.append((_min, _max))

    def add_image(self, image_dat):
        """Add a new image to VersionedInstance from raw data.
        """
        name = image_dat.get('name')
        variable = image_dat.get('variable')
        role = image_dat.get('role', None)
        roles = image_dat.get('roles', None)
        if variable not in self._images:
            self._images[variable] = {
                'name': name,
                'variable': variable,
                'role': role,
                'roles': roles
            }
        else:
            raise ValueError(
                'Duplicated image variables definitions for %s %s' % (
                    self.instance_type, self.major_version
                ))

    def add_release(self, release_dat):
        """Add a new release to VersionedInstance from raw data.
        """
        r = Release(self.instance_type, release_dat)
        # Validate the image completeness
        for image_name in r.image_version:
            assert image_name in self._images, \
                'Image name %s of release %s %s should be declared first' % (
                    image_name, self.instance_type, self.major_version)
        self._releases[r.release_version] = r

    def get_release(self, release_version, default=None):
        """Get a release defined in the VersionedInstance
        """
        release_version = parse_version(release_version)
        return self._releases.get(release_version, default)

    def has_release(self, release_version):
        """Check if a specific version is defined in the VersionedInstance
        """
        release_version = parse_version(release_version)
        return release_version in self._releases

    def create_release(self, version, from_release=None, with_major=False):
        """Create a new release with version from @from_release or
        """
        version = parse_version(version)
        if version in self._releases:
            print('Warning: Duplicated new version {} for {} {}, skip'.format(
                version, self.instance_type, self.major_version
            ))
            return

        if from_release is None:
            from_release = self.find_latest_release(product_name(version))
            assert from_release is not None, \
                'No valid final version found in {}, {} as reference to create {}'.format(
                    self.instance_type, self.major_version, version
                )

        new_release = from_release.clone_as(version)
        self._releases[version] = new_release

        if with_major:
            major_version = to_major_version(version)
            if major_version not in self._releases:
                minor_release = new_release.clone_as(major_version)
                minor_release.is_final = False
                self._releases[major_version] = minor_release
            else:
                print('Duplicated major version {} for {}, {}, skip'.format(
                    major_version, self.instance_type, self.major_version))

    def find_latest_release(self, product=None, is_final=False):
        """Find the latest release by product name .
        """
        latest_release = None
        for r in self.ordered_releases[::-1]:
            # Ignore non-final releases
            if is_final and not r.is_final:
                continue

            # Find the first release (reversely) for given product
            rp = product_name(r.release_version)
            if product is not None:
                latest_release = r if rp == product else None
            else:
                latest_release = r if rp != product else None

            if latest_release is not None:
                break

        return latest_release

    def validate(self, release_meta):
        """Validate properties and fix errors if possible
        """
        # Update tdc min-max versions
        self.update_tdc_minmax_version(release_meta)

        # Validate each release
        for ver, release in self._releases.items():
            release.validate_final_flag()
            release.validate_tdc_minmax_version(self._min_tdc_version, self._max_tdc_version)

        self._validate_hot_fix_ranges()
        # self._validate_tdc_not_dependent_on_other_product_lines()  # Disable it for now
        self._validate_releases(release_meta)
        self._validate_argodb_images(release_meta)

    def update_tdc_minmax_version(self, release_meta):
        global_range = release_meta.get_tdc_version_range()
        tdc_vranges = list()
        for release in self.ordered_releases:
            # The third party release (without version prefix) takes the global version range
            if release.is_third_party() and global_range is not None:
                tdc_vranges.append(global_range)
                continue

            # Otherwise get the precise compatible product version range
            vrange = release_meta.get_tdc_version_range(release.release_version)
            if vrange is not None:
                tdc_vranges.append(vrange)
            else:
                print('Warning: not found a valid tdc version range for {}, {}. Use the global range instead'
                      .format(release.instance_type, release.release_version))
                if global_range is not None:
                    tdc_vranges.append(global_range)

        if len(tdc_vranges) > 0:
            self._min_tdc_version = sorted([i[0] for i in tdc_vranges], key=cmp_to_key(
                lambda x, y: x.compares(y)
            ))[0]
            self._max_tdc_version = sorted([i[1] for i in tdc_vranges], key=cmp_to_key(
                lambda x, y: x.compares(y)
            ))[-1]
        else:
            raise ValueError('At least a valid release is required for {}, {}'.format(
                self.instance_type, self.major_version)
            )

    def _validate_hot_fix_ranges(self):
        """Validate release versions and hot-fix ranges.
        Fix errors if possible
        """
        for release in self._releases.values():
            v = release.release_version
            try:
                self._is_version_in_hot_fix_range(v)
            except AssertionError as e:
                print('Warning {}'.format(str(e)))
                self.add_hot_fix_range(v, v)

        # Merge continuous hot-fix ranges
        # Differentiate complete and minor-versioned-only versions
        complete_ranges = list()
        major_versions = list()
        for minv, maxv in self._hot_fix_ranges:
            im_minv = is_major_version(minv)
            im_maxv = is_major_version(maxv)
            assert im_minv == im_maxv, 'Min and max should take the same form'
            if im_minv:
                major_versions.append((minv, maxv))
            else:
                complete_ranges.append((minv, maxv))
        self._hot_fix_ranges = concatenate_vranges(complete_ranges) + \
                               concatenate_vranges(major_versions)

    def _is_version_in_hot_fix_range(self, version):
        found = False
        for _min, _max in self._hot_fix_ranges:
            if version.in_range(_min, _max):
                found = True
                break
        assert found is True, \
            'Release version %s of "%s" should be located in a specific hot-fix range' % \
            (version, self.instance_type)

    def _validate_tdc_not_dependent_on_other_product_lines(self):
        for release in self._releases.values():
            product = product_name(release.release_version)
            if product == VC.OEM_NAME:
                for dep, (minv, maxv) in release.dependencies.items():
                    if product_name(minv) != product:
                        print('Warning: TDC should be independent product, instance "{}", {}, dep "{}"'
                              .format(release.instance_type, release.release_version, dep))

    def _validate_releases(self, release_meta):
        """Validate all releases of the instance.
        """
        ## For debugging
        # if self.instance_type == 'sophon' and str(self.major_version) == '2.3':
        #     print(self.instance_type, self.major_version)

        for r in self.ordered_releases:
            # Get compatible version ranges for each product
            #   {product: [(minv, maxv), (minv, maxv)]}
            # WARP-34008: support instance-specific constraints
            cv = release_meta.get_compatible_versions(r.release_version, instance_name=r.instance_type)

            # Filter vrange by tdc min-max version
            _is_major_version = is_major_version(r.release_version)
            minv = parse_version(self._min_tdc_version, _is_major_version)
            maxv = parse_version(self._max_tdc_version, _is_major_version)
            if product_name(r.release_version) == VC.OEM_NAME:
                for pname in cv:
                    filtered = list()
                    for v in cv[pname]:
                        fv = filter_vrange(v, (minv, maxv))
                        if fv is not None:
                            filtered.append(fv)
                    if not filtered:
                        print('Warning: Release {} of instance "{}" is filtered out by min-max tdc version.'
                              .format(r.release_version, r.instance_type))
                    cv[pname] = filtered

            # Validate the dependency versions
            for instance, vrange in r.dependencies.items():
                product = product_name(vrange[0])
                if product not in cv:
                    minv, maxv = vrange
                else:
                    if len(cv[product]) == 0:
                        raise ValueError(
                            'No valid version range declared for instance {}, version {} in releasemeta'
                                .format(self.instance_type, r.release_version))
                    else:
                        # TODO: it seems safe to merge dependency release versions
                        minv, maxv = concatenate_vranges(cv[product], hard_merging=True)[0]
                if minv != vrange[0]:
                    print('Warning: incompatible min version {} (should be {}) for dep "{}" of release "{}" version {}'
                          .format(vrange[0], minv, instance, r.instance_type, r.release_version))
                if maxv != vrange[1]:
                    print('Warning: incompatible max version {} (should be {}) for dep "{}" of release "{}" version {}'
                          .format(vrange[1], maxv, instance, r.instance_type, r.release_version))
                r.dependencies[instance] = (minv, maxv)

    def _validate_argodb_images(self, release_meta):
        """
        Fix custom terminal image version for argodb, WARP-38405
        """
        if self.instance_type == 'terminal':
            for release_ver, release in self._releases.items():
                if release_ver.prefix != 'argodb':
                    continue
                tdc_vrange = release_meta.get_tdc_version_range(release_ver, self.instance_type)
                release.image_version['terminal_image'] = tdc_vrange[1]
                print('WARNING: set terminal image for argodb {} as {} (WARP-38405)'.format(release_ver, tdc_vrange[1]))

    def to_yaml(self):
        # Ordered keys
        res = OrderedDict()
        res['instance-type'] = self.instance_type
        res['major-version'] = str(self.major_version)
        res['min-tdc-version'] = str(self.min_tdc_version)
        res['max-tdc-version'] = str(self.max_tdc_version)
        res['hot-fix-ranges'] = list()
        for vrange in self.hot_fix_ranges:
            res['hot-fix-ranges'].append({
                'max': str(vrange[1]),
                'min': str(vrange[0])
            })

        res['images'] = list()
        for var, data in self.images:
            image_dat = {key: value for key, value in data.items() if value is not None}
            res['images'].append(image_dat)

        res['releases'] = list()
        for r in self.ordered_releases:
            robj = OrderedDict()
            robj['release-version'] = str(r.release_version)

            robj['image-version'] = dict()
            for img_name, ver in r.image_version.items():
                robj['image-version'][img_name] = str(ver)
            robj['dependencies'] = list()
            for instance, (minv, maxv) in r.dependencies.items():
                robj['dependencies'].append({
                    'max-version': str(maxv),
                    'min-version': str(minv),
                    'type': instance
                })
            robj['final'] = r.is_final
            res['releases'].append(robj)

        return ordered_yaml_dump(res, default_flow_style=False)

    def convert_oem(self):
        self._min_tdc_version = replace_product_name(self._min_tdc_version, VC.OEM_NAME, VC._OEM_ORIGIN)
        self._max_tdc_version = replace_product_name(self._max_tdc_version, VC.OEM_NAME, VC._OEM_ORIGIN)
        self._hot_fix_ranges = [
            (
                replace_product_name(minv, VC.OEM_NAME, by=VC._OEM_ORIGIN),
                replace_product_name(maxv, VC.OEM_NAME, by=VC._OEM_ORIGIN)
            ) for minv, maxv in self._hot_fix_ranges
        ]
        for rver, release in self._releases.items():
            self._releases[rver] = release.convert_oem(VC.OEM_NAME, VC._OEM_ORIGIN)


class Release(object):
    """ The metadata of a specific versioned release.
    """

    def __init__(self, instance_type, val):
        self.instance_type = instance_type
        self.release_version = parse_version(val.get('release-version'))
        self.is_final = val.get('final', False)

        # Images
        self.image_version = dict()  # {image_var: version}
        for img, ver in val.get('image-version', dict()).items():
            self.image_version[img] = parse_version(ver)

        # Dependencies
        self.dependencies = dict()  # {instance_type: (minv, maxv)}
        for dep in val.get('dependencies', list):
            instance_type = dep.get('type')
            _max_ver = parse_version(dep.get('max-version'))
            _min_ver = parse_version(dep.get('min-version'))

            if _min_ver.compares(_max_ver) > 0:
                raise ValueError('Invalid min-max version declaim for dependency of %s for %s %s'
                                 % (instance_type,
                                    self.instance_type,
                                    self.release_version)
                                 )

            if instance_type in self.dependencies:
                raise ValueError('Duplicated dependency of %s for %s %s' %
                                 (instance_type, self.instance_type,
                                  self.release_version))
            else:
                self.dependencies[instance_type] = (_min_ver, _max_ver)

    def convert_oem(self, oemname, by=VC._OEM_ORIGIN):
        self.release_version = replace_product_name(self.release_version, oemname, by)
        for image_name, ver in self.image_version.items():
            self.image_version[image_name] = replace_product_name(ver, oemname, by)
        for instance, vrange in self.dependencies.items():
            minv, maxv = vrange
            self.dependencies[instance] = (
                replace_product_name(minv, oemname, by),
                replace_product_name(maxv, oemname, by)
            )
        return self

    def clone_as(self, version):
        """Clone a new versioned release with reference to self.
        """
        version = parse_version(version)
        _is_major_version = is_major_version(version)
        new_release = copy.deepcopy(self)
        new_release.release_version = version
        new_release.is_final = not _is_major_version
        for img, ver in new_release.image_version.items():
            if product_name(ver) == product_name(self.release_version):
                new_release.image_version[img] = version
            elif _is_major_version:
                new_release.image_version[img] = to_major_version(ver)
        for dep, (minv, maxv) in new_release.dependencies.items():
            if product_name(minv) == product_name(self.release_version):
                new_release.dependencies[dep] = (version, version)
            elif _is_major_version:
                new_release.dependencies[dep] = (to_major_version(minv), to_major_version(maxv))
        return new_release

    def validate_tdc_minmax_version(self, minv, maxv):
        """Validate the release version should fall into min-max tdc range.
        """
        if product_name(self.release_version) == VC.OEM_NAME \
                and self.release_version.suffix is not None:
            assert self.release_version.in_range(minv, maxv), \
                'The release {} of "{}" should in min-max tdc versions ({}, {})'.format(
                    self.release_version, self.instance_type,
                    minv, maxv
                )

    def is_major_version(self):
        return is_major_version(self.release_version)

    def is_third_party(self):
        # Third-party images without product name
        # e.g., weblogic, nexus
        product = product_name(self.release_version)
        if product is None and self.release_version.suffix is None:
            return True
        return False

    def validate_final_flag(self):
        """Validate is_final flag and version format
        """
        is_valid_final = False
        if self.release_version.suffix is not None or self.is_third_party():
            is_valid_final = True
        if self.is_final and not is_valid_final:
            raise ValueError('The final version {} of instance {} is illegal'.format(
                self.release_version, self.instance_type
            ))



# Copyright (c) 2018-2019 `Transwarp, Inc. <http://www.transwarp.io>`_.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the Transwarp, Inc nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import os

from setuptools import setup, find_packages

__version__ = '1.3.4.2'


def walk_path_files(directory, target_folder=None):
    res = {}
    for subdir, dir, files in os.walk(directory):
        if target_folder is not None:
            target = os.path.join(target_folder, subdir)
        else:
            target = subdir
        if target not in res:
            res[target] = []
        for fname in files:
            res[target].append(os.path.join(subdir, fname))
    return res


data_files = []

requirements = open(os.path.join(os.path.dirname(__file__), 'requirements.txt')).readlines()
install_requires = [i.strip('\r\n ') for i in requirements]

setup(
    name='verminator',
    version=__version__,
    packages=find_packages(),
    include_package_data=True,
    data_files=data_files,
    url='http://github.com/WarpCloud/verminator',
    license='Transwarp',
    author='Xiaming Chen',
    author_email='xiaming.chen@transwarp.io',
    description='TDC image Version management tERMINATOR.',
    install_requires=install_requires,
    scripts=[
        'bin/verminator'
    ],
    keywords=['utility', 'versioning'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



import copy
from collections import OrderedDict
from functools import cmp_to_key

import yaml
from flex_version import FlexVersion, VersionMeta, VersionDelta

# Customized version suffix ordering
FlexVersion.ordered_suffix = ['rc', 'final', None]
VersionMeta._v_regex = r"(?P<prefix>.*\-)?(?P<major>\d+)(?P<minor>\.\d+)?" + \
               r"(?P<maintenance>\.\d+)?(?P<build>\.\d+)?(?P<suffix_raw>\-.*)?"


def ordered_yaml_load(yaml_path, Loader=yaml.Loader,
                      object_pairs_hook=OrderedDict):
    class OrderedLoader(Loader):
        pass

    def construct_mapping(loader, node):
        loader.flatten_mapping(node)
        return object_pairs_hook(loader.construct_pairs(node))

    OrderedLoader.add_constructor(
        yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
        construct_mapping)
    with open(yaml_path) as stream:
        return yaml.load(stream, OrderedLoader)


def ordered_yaml_dump(data, stream=None, Dumper=yaml.SafeDumper, **kwds):
    class OrderedDumper(Dumper):
        pass

    def _dict_representer(dumper, data):
        return dumper.represent_mapping(
            yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
            data.items())

    OrderedDumper.add_representer(OrderedDict, _dict_representer)
    return yaml.dump(data, stream, OrderedDumper, **kwds)


def parse_version(version, major_versioned=False):
    if version is None:
        return None

    if not isinstance(version, VersionMeta):
        version = FlexVersion.parse_version(version)
    else:
        version = copy.deepcopy(version)

    if major_versioned:
        version.maintenance = None
        version.build = None
        version.suffix = None
        version.suffix_version = None

    return version


def product_name(version):
    version = parse_version(version)
    return version.prefix


def replace_product_name(version, newname, by=None):
    version = parse_version(version)
    if by is not None and version.prefix == by:
        version.prefix = newname
    elif by is None:
        version.prefix = newname
    return version


def is_major_version(version):
    version = parse_version(version)
    return version.maintenance is None \
           and version.build is None \
           and version.suffix is None \
           and version.suffix_version is None


def to_major_version(version):
    version = parse_version(version)
    version.maintenance = None
    version.build = None
    version.suffix = None
    version.suffix_version = None
    return version


def filter_vrange(this, other):
    """Filter version range of `this` against the `other`.
    """
    tmin, tmax = this
    omin, omax = other

    rmin, rmax = tmin, tmax
    if product_name(tmin) == product_name(omin):
        rmin = tmin if omin <= tmin else omin if omin <= tmax else None
        rmax = tmax if omax >= tmax else omax if omax >= tmin else None

    if None in (rmin, rmax):
        return None
    else:
        return (rmin, rmax)


def concatenate_vranges(vranges, hard_merging=False):
    """
    Connect and merge version ranges.

    :param vranges: a list of version ranges, i.e. [(minv, maxv), ...]
    :param hard_merging: if performing merging without considering range overlapping.
    :return: a list of concatenated version ranges, i.e. [(minv, maxv), ...]
    """
    if len(vranges) <= 1:
        return vranges

    prefixes = dict()
    for vrange in vranges:
        prefix = vrange[0].prefix
        if prefix not in prefixes:
            prefixes[prefix] = list()
        prefixes[prefix].append(vrange)

    concatenated = [
        _concatenate_vranges_with_same_prefix(vranges, hard_merging)
        for vranges in prefixes.values()
    ]

    return [y for x in concatenated for y in x]


def _concatenate_vranges_with_same_prefix(vranges, hard_merging=False):
    """
    Connect and merge version ranges.

    :param vranges: a list of version ranges, i.e. [(minv, maxv), ...]
    :param hard_merging: if performing merging without considering range overlapping.
    :return: a list of concatenated version ranges, i.e. [(minv, maxv), ...]
    """
    sorted_vranges = sorted(vranges, key=cmp_to_key(
        lambda x, y: FlexVersion.compares(x[0], y[0])
    ))

    res = [sorted_vranges[0]]
    for vrange in sorted_vranges[1:]:
        pmin, pmax = res[-1]
        cmin, cmax = vrange

        if not hard_merging:
            # Ranges connected directly:
            # * Overlapping ranges
            # * adjacent suffix versions
            if cmin.suffix is not None and pmin.suffix is not None:
                if cmin == pmax.add(VersionDelta(sver=1)):
                    res[-1] = (pmin, cmax)
                    continue
                else:
                    overlapped = pmin.in_range(cmin, cmax) \
                                 or pmax.in_range(cmin, cmax) \
                                 or cmin.in_range(pmin, pmax) \
                                 or cmax.in_range(pmin, pmax)
                    if overlapped:
                        minv = cmin if cmin < pmin else pmin
                        maxv = cmax if cmax > pmax else pmax
                        res[-1] = (minv, maxv)
                        continue

            # Ranges between rc and final
            if pmax.substitute(cmin, ignore_suffix=True) == VersionDelta.zero \
                    and pmax.suffix == 'rc' and cmin.suffix == 'final':
                res[-1] = (pmin, cmax)
                continue

            # Ranges between final and rc
            delta = cmin.substitute(pmax, ignore_suffix=True)
            if delta >= VersionDelta.zero and pmax.suffix == 'final':
                res[-1] = (pmin, cmax)
                continue

            # Strict concatenation policy
            res.append(vrange)
        else:
            res[-1] = (pmin, cmax)

    return res
